# Copyright (C) 2023-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta, timezone
import itertools
import logging
import operator
import shutil
import subprocess
import textwrap
from typing import Any, List, Set
from unittest.mock import call

import pytest
import yaml

from swh.model.model import BaseModel, Origin
from swh.model.swhids import CoreSWHID, ExtendedObjectType, ExtendedSWHID
from swh.objstorage.interface import ObjStorageInterface
from swh.search.interface import SearchInterface
from swh.storage.interface import StorageInterface
from swh.storage.proxies.masking.db import MaskedState

from ..mirror_notification_watcher import MASKING_REQUEST_IDENTIFIER_PREFIX
from ..operations import MaskingRequestNotFound, Removable, Remover, RemoverError
from ..recovery_bundle import SecretSharing
from .conftest import (
    OBJECT_SECRET_KEY,
    TWO_GROUPS_REQUIRED_WITH_ONE_MINIMUM_SHARE_EACH_SECRET_SHARING_YAML,
)


@pytest.fixture
def remover(
    sample_populated_storage,
    graph_client_with_only_initial_origin,
):
    return Remover(
        storage=sample_populated_storage,
        graph_client=graph_client_with_only_initial_origin,
        known_missing=[
            ExtendedSWHID.from_string(
                "swh:1:cnt:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
            )
        ],
    )


def test_remover_get_removable(remover):
    swhids = [
        ExtendedSWHID.from_string("swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054"),
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165"),
    ]
    removable = remover.get_removable(swhids)
    assert len(removable.removable_swhids) == 33
    assert len(removable.referencing) == 0


def test_remover_get_removable_populates_referencing(remover):
    swhids = [
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165"),
    ]
    removable = remover.get_removable(swhids)
    assert set(removable.referencing) == {
        ExtendedSWHID.from_string(s)
        for s in (
            "swh:1:rel:0000000000000000000000000000000000000010",
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:dir:0000000000000000000000000000000000000008",
        )
    }


def test_remover_get_removable_with_known_missing(
    sample_populated_storage,
    graph_client_with_both_origins,
):
    swhids = [
        ExtendedSWHID.from_string("swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054"),
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165"),
    ]
    remover = Remover(
        storage=sample_populated_storage,
        graph_client=graph_client_with_both_origins,
        known_missing=swhids[1:],
    )
    removable = remover.get_removable(swhids)
    assert removable.removable_swhids == [
        ExtendedSWHID.from_string(s)
        for s in (
            "swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054",
            "swh:1:snp:0000000000000000000000000000000000000020",
            "swh:1:emd:ba1e287385aac8d76caaf9956819a5d68bfe2083",
            "swh:1:emd:bfe476f7cffb00a5be2b12cfb364e207e4be0da2",
            "swh:1:emd:1ecd328c7597043895621da4d5351c59f1de663c",
            "swh:1:emd:bcfe01c5e96a675b500d32b15b4ea36bd5a46cdb",
        )
    ]
    assert set(removable.referencing) == {
        ExtendedSWHID.from_string(s)
        for s in (
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:rel:0000000000000000000000000000000000000010",
        )
    }


@pytest.mark.skipif(
    not shutil.which("gc"), reason="missing `gc` executable from graphviz"
)
def test_remover_output_inventory_subgraph(tmp_path, remover):
    swhids = [
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165")
    ]
    dot_path = tmp_path / "subgraph.dot"
    _ = remover.get_removable(swhids, output_inventory_subgraph=dot_path.open("w"))
    completed_process = subprocess.run(
        ["gc", dot_path],
        check=True,
        capture_output=True,
    )
    assert b"      21      24 Inventory" in completed_process.stdout


@pytest.mark.skipif(
    not shutil.which("gc"), reason="missing `gc` executable from graphviz"
)
def test_remover_output_removable_subgraph(tmp_path, remover):
    swhids = [
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165")
    ]
    dot_path = tmp_path / "subgraph.dot"
    _ = remover.get_removable(swhids, output_removable_subgraph=dot_path.open("w"))
    completed_process = subprocess.run(
        ["gc", dot_path],
        check=True,
        capture_output=True,
    )
    assert b"      21      24 Removable" in completed_process.stdout


@pytest.mark.skipif(
    not shutil.which("gc"), reason="missing `gc` executable from graphviz"
)
def test_remover_output_pruned_removable_subgraph(tmp_path, remover):
    swhids = [
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165")
    ]
    dot_path = tmp_path / "subgraph.dot"
    _ = remover.get_removable(
        swhids, output_pruned_removable_subgraph=dot_path.open("w")
    )
    completed_process = subprocess.run(
        ["gc", dot_path],
        check=True,
        capture_output=True,
    )
    assert b"      11      10 Removable" in completed_process.stdout


@pytest.fixture
def secret_sharing_conf():
    return yaml.safe_load(
        TWO_GROUPS_REQUIRED_WITH_ONE_MINIMUM_SHARE_EACH_SECRET_SHARING_YAML
    )["secret_sharing"]


def test_remover_create_recovery_bundle(
    remover,
    secret_sharing_conf,
    tmp_path,
    sample_populated_storage,
):
    swhids = [
        ExtendedSWHID.from_string(s)
        for s in (
            "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
            "swh:1:snp:0000000000000000000000000000000000000022",
            "swh:1:rel:0000000000000000000000000000000000000021",
            "swh:1:rev:0000000000000000000000000000000000000018",
            "swh:1:rev:0000000000000000000000000000000000000013",
            "swh:1:dir:0000000000000000000000000000000000000017",
            "swh:1:cnt:0000000000000000000000000000000000000015",
            "swh:1:cnt:0000000000000000000000000000000000000014",
        )
    ]
    referencing = [
        ExtendedSWHID.from_string(s)
        for s in (
            "swh:1:rel:0000000000000000000000000000000000000010",
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:dir:0000000000000000000000000000000000000012",
        )
    ]
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    expire = datetime.now(timezone.utc) + timedelta(days=365)
    share_ids = {
        share_id
        for group in secret_sharing_conf["groups"].values()
        for share_id in group["recipient_keys"].keys()
    }
    remover.create_recovery_bundle(
        secret_sharing=SecretSharing.from_dict(secret_sharing_conf),
        requested=[swhids[0]],
        removable=Removable(removable_swhids=swhids, referencing=referencing),
        recovery_bundle_path=bundle_path,
        removal_identifier="test",
        reason="doing a test",
        expire=expire,
    )

    from ..recovery_bundle import RecoveryBundle

    bundle = RecoveryBundle(bundle_path)
    assert len(bundle.swhids) == len(swhids)
    assert bundle.removal_identifier == "test"
    assert bundle.reason == "doing a test"
    assert bundle.expire.isoformat(timespec="seconds") == expire.isoformat(
        timespec="seconds"
    )
    assert bundle.share_ids == share_ids


def test_remover_create_recovery_bundle_fails_with_expire_in_the_past(
    remover,
    secret_sharing_conf,
    tmp_path,
):
    swhids = [
        ExtendedSWHID.from_string(s)
        for s in ("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",)
    ]
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    expire = datetime.fromisoformat("2001-01-01").astimezone()
    with pytest.raises(RemoverError, match="Unable to set expiration date"):
        remover.create_recovery_bundle(
            secret_sharing=SecretSharing.from_dict(secret_sharing_conf),
            requested=[swhids[0]],
            removable=Removable(removable_swhids=swhids, referencing=[]),
            recovery_bundle_path=bundle_path,
            removal_identifier="test",
            reason="doing a test",
            expire=expire,
        )


def test_remover_remove(
    mocker,
    sample_populated_storage,
    graph_client_with_only_initial_origin,
):
    removal_storage_one = mocker.MagicMock()
    removal_storage_one.object_delete.return_value = {"origin:delete": 0}
    removal_storage_one.extid_delete_for_target.return_value = {"extid:delete": 0}
    removal_storage_two = mocker.MagicMock()
    removal_storage_two.object_delete.return_value = {"origin:delete": 0}
    removal_storage_two.extid_delete_for_target.return_value = {"extid:delete": 0}
    remover = Remover(
        sample_populated_storage,
        graph_client_with_only_initial_origin,
        removal_storages={"one": removal_storage_one, "two": removal_storage_two},
    )
    core_swhids = """\
        swh:1:snp:0000000000000000000000000000000000000022
        swh:1:rev:0000000000000000000000000000000000000018
        swh:1:rel:0000000000000000000000000000000000000021
    """
    remover.swhids_to_remove = [
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165")
    ] + [
        ExtendedSWHID.from_string(line.strip())
        for line in core_swhids.rstrip().splitlines()
    ]
    remover.remove()
    for storage in (removal_storage_one, removal_storage_two):
        storage.object_delete.assert_called_once()
        args, _ = storage.object_delete.call_args
        assert set(args[0]) == set(remover.swhids_to_remove)
        storage.extid_delete_for_target.assert_called_once()
        args, _ = storage.extid_delete_for_target.call_args
        assert set(args[0]) == {
            CoreSWHID.from_string(line.strip())
            for line in core_swhids.rstrip().splitlines()
        }


def test_remover_remove_from_objstorages(
    mocker,
    sample_populated_storage,
):
    from swh.objstorage.interface import objid_from_dict

    storage = sample_populated_storage
    objstorage1 = mocker.Mock(spec=ObjStorageInterface)
    objstorage2 = mocker.Mock(spec=ObjStorageInterface)
    graph_client = mocker.MagicMock()
    remover = Remover(
        storage,
        graph_client,
        removal_objstorages={"one": objstorage1, "two": objstorage2},
    )
    remover.swhids_to_remove = [
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165"),
    ]
    contents = storage.content_get(
        [bytes.fromhex("0000000000000000000000000000000000000014")], algo="sha1_git"
    )
    remover.objids_to_remove = [
        objid_from_dict(content.to_dict()) for content in contents
    ]
    remover.remove()
    for objstorage in (objstorage1, objstorage2):
        objstorage.delete.assert_called_once()


def test_remover_remove_from_flaping_objstorages(
    mocker,
    sample_populated_storage,
):
    from swh.objstorage.interface import objid_from_dict

    storage = sample_populated_storage
    mocker.patch("swh.alter.operations.time.sleep")
    objstorage = mocker.Mock(spec=ObjStorageInterface)
    objstorage.delete.side_effect = itertools.cycle([Exception(), Exception(), None])
    graph_client = mocker.MagicMock()
    remover = Remover(
        storage,
        graph_client,
        removal_objstorages={"flaping": objstorage},
    )
    remover.swhids_to_remove = [
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165"),
    ]
    contents = storage.content_get(
        [bytes.fromhex("0000000000000000000000000000000000000014")], algo="sha1_git"
    )
    remover.objids_to_remove = [
        objid_from_dict(content.to_dict()) for content in contents
    ]
    remover.remove()
    assert len(objstorage.delete.call_args_list) == 3


def test_remover_remove_from_faulty_objstorages(
    mocker,
    caplog,
    sample_populated_storage,
):
    from swh.objstorage.interface import objid_from_dict

    storage = sample_populated_storage
    mocker.patch("swh.alter.operations.time.sleep")
    objstorage = mocker.Mock(spec=ObjStorageInterface)
    objstorage.delete.side_effect = Exception("error")
    graph_client = mocker.MagicMock()
    remover = Remover(
        storage,
        graph_client,
        removal_objstorages={
            "faulty": objstorage,
        },
    )
    remover.swhids_to_remove = [
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165"),
    ]
    contents = storage.content_get(
        [bytes.fromhex("0000000000000000000000000000000000000014")], algo="sha1_git"
    )
    remover.objids_to_remove = [
        objid_from_dict(content.to_dict()) for content in contents
    ]
    with caplog.at_level(logging.WARNING):
        with pytest.raises(Exception, match="error"):
            remover.remove()
    assert len(objstorage.delete.call_args_list) == 3
    assert (
        "objstorage “faulty” raised “Exception('error')” during attempt 2, "
        "retrying in 10 seconds…" in caplog.text
    )


def test_remover_remove_from_objstorages_object_missing_from_one_objstorage(
    caplog,
    mocker,
    sample_populated_storage,
):
    from swh.objstorage.exc import ObjNotFoundError
    from swh.objstorage.interface import objid_from_dict

    storage = sample_populated_storage
    objstorage1 = mocker.Mock(spec=ObjStorageInterface)
    objstorage2 = mocker.Mock(spec=ObjStorageInterface)
    objstorage2.delete.side_effect = ObjNotFoundError()
    graph_client = mocker.MagicMock()
    remover = Remover(
        storage,
        graph_client,
        removal_objstorages={"one": objstorage1, "two": objstorage2},
    )
    contents = storage.content_get(
        [bytes.fromhex("0000000000000000000000000000000000000014")], algo="sha1_git"
    )
    remover.objids_to_remove = [
        objid_from_dict(content.to_dict()) for content in contents
    ]
    with caplog.at_level(logging.INFO):
        remover.remove_from_objstorages()
    # We should not get an error message for a single objstorage.
    # Rationale: more objstorages have been added as time went
    # on. But while old objects have not been moved to newer
    # objstorages, they are still available. The real problem
    # is when not a single objstorage has a particular object.
    # (See the test below.)
    assert not any("not found" in msg for msg in caplog.messages)


def test_remover_remove_from_objstorages_object_missing_from_all_objstorages(
    caplog,
    mocker,
    sample_populated_storage,
):
    from swh.objstorage.exc import ObjNotFoundError
    from swh.objstorage.interface import objid_from_dict

    storage = sample_populated_storage
    objstorage1 = mocker.Mock(spec=ObjStorageInterface)
    objstorage1.delete.side_effect = ObjNotFoundError()
    objstorage2 = mocker.Mock(spec=ObjStorageInterface)
    objstorage2.delete.side_effect = ObjNotFoundError()
    graph_client = mocker.MagicMock()
    remover = Remover(
        storage,
        graph_client,
        removal_objstorages={"one": objstorage1, "two": objstorage2},
    )
    contents = storage.content_get(
        [bytes.fromhex("0000000000000000000000000000000000000014")], algo="sha1_git"
    )
    remover.objids_to_remove = [
        objid_from_dict(content.to_dict()) for content in contents
    ]
    with caplog.at_level(logging.INFO):
        remover.remove_from_objstorages()
    expected = textwrap.dedent(
        """\
        Objects not found in any objstorage:
        | blake2s256                                                       | sha1                                     | sha1_git                                 | sha256                                                           |
        |------------------------------------------------------------------|------------------------------------------|------------------------------------------|------------------------------------------------------------------|
        | 0000000000000000000000000000000000000000000000000000000000000014 | 0000000000000000000000000000000000000014 | 0000000000000000000000000000000000000014 | 0000000000000000000000000000000000000000000000000000000000000014 |
        """.rstrip()  # noqa: B950
    )
    assert expected == caplog.records[-1].message


def test_remover_remove_from_objstorage_display_timings(
    caplog,
    mocker,
    sample_populated_storage,
):
    from swh.objstorage.interface import objid_from_dict

    storage = sample_populated_storage
    objstorage = mocker.Mock(spec=ObjStorageInterface)
    graph_client = mocker.MagicMock()
    remover = Remover(
        storage,
        graph_client,
        removal_objstorages={"objstorage": objstorage},
    )
    remover.swhids_to_remove = [
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165"),
    ]
    contents = storage.content_get(
        [
            bytes.fromhex("0000000000000000000000000000000000000007"),
            bytes.fromhex("0000000000000000000000000000000000000011"),
            bytes.fromhex("0000000000000000000000000000000000000014"),
        ],
        algo="sha1_git",
    )
    mocker.patch("time.monotonic", side_effect=[0.001, 0.2, 0.3, 0.4, 0.5, 0.72])
    remover.objids_to_remove = [
        objid_from_dict(content.to_dict()) for content in contents
    ]
    with caplog.at_level(logging.INFO):
        remover.remove()
    assert caplog.messages[-2] == (
        "3 objects removed from objstorage “objstorage”. "
        "Total time: 519 milliseconds, "
        "average: 173 milliseconds per object, "
        "standard deviation: 64.09 milliseconds"
    )


def test_remover_remove_from_searches(
    mocker,
    sample_populated_storage,
):
    storage = sample_populated_storage
    search1 = mocker.Mock(spec=SearchInterface)
    search2 = mocker.Mock(spec=SearchInterface)
    graph_client = mocker.MagicMock()
    remover = Remover(
        storage,
        graph_client,
        removal_searches={"one": search1, "two": search2},
    )
    remover.origin_urls_to_remove = [
        "https://example.com/swh/graph1",
        "https://example.com/swh/graph2",
    ]
    remover.remove()
    for search in (search1, search2):
        assert search.origin_delete.call_args_list == [
            call("https://example.com/swh/graph1"),
            call("https://example.com/swh/graph2"),
        ]
        search.flush.assert_called_once()


def test_remover_have_new_references_outside_removed(
    mocker,
    sample_populated_storage,
    remover,
):
    storage = sample_populated_storage
    swhids = [
        "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
        "swh:1:snp:0000000000000000000000000000000000000022",
        "swh:1:rel:0000000000000000000000000000000000000021",
        "swh:1:rev:0000000000000000000000000000000000000018",
        "swh:1:rev:0000000000000000000000000000000000000013",
        "swh:1:dir:0000000000000000000000000000000000000017",
        "swh:1:cnt:0000000000000000000000000000000000000015",
        "swh:1:cnt:0000000000000000000000000000000000000014",
    ]
    mocker.patch.object(
        storage,
        "object_find_recent_references",
        wraps=lambda s, _: (
            [
                ExtendedSWHID.from_string(
                    "swh:1:rev:0000000000000000000000000000000000000003"
                )
            ]
            if s.object_type == ExtendedObjectType.DIRECTORY
            else []
        ),
    )
    result = remover.have_new_references(
        [ExtendedSWHID.from_string(swhid) for swhid in swhids]
    )
    assert result is True


def test_remover_have_new_references_inside_removed(
    mocker,
    sample_populated_storage,
    remover,
):
    storage = sample_populated_storage
    swhids = [
        "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
        "swh:1:snp:0000000000000000000000000000000000000022",
        "swh:1:rel:0000000000000000000000000000000000000021",
        "swh:1:rev:0000000000000000000000000000000000000018",
        "swh:1:rev:0000000000000000000000000000000000000013",
        "swh:1:dir:0000000000000000000000000000000000000017",
        "swh:1:cnt:0000000000000000000000000000000000000015",
        "swh:1:cnt:0000000000000000000000000000000000000014",
    ]
    mocker.patch.object(
        storage,
        "object_find_recent_references",
        wraps=lambda s, _: (
            [
                ExtendedSWHID.from_string(
                    "swh:1:rev:0000000000000000000000000000000000000013"
                )
            ]
            if s.object_type == ExtendedObjectType.DIRECTORY
            else []
        ),
    )
    result = remover.have_new_references(
        [ExtendedSWHID.from_string(swhid) for swhid in swhids]
    )
    assert result is False


def test_remover_have_new_references_nothing_new(
    mocker,
    sample_populated_storage,
    remover,
):
    storage = sample_populated_storage
    swhids = [
        "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
        "swh:1:snp:0000000000000000000000000000000000000022",
        "swh:1:rel:0000000000000000000000000000000000000021",
        "swh:1:rev:0000000000000000000000000000000000000018",
        "swh:1:rev:0000000000000000000000000000000000000013",
        "swh:1:dir:0000000000000000000000000000000000000017",
        "swh:1:cnt:0000000000000000000000000000000000000015",
        "swh:1:cnt:0000000000000000000000000000000000000014",
    ]
    mocker.patch.object(storage, "object_find_recent_references", return_value=[])
    result = remover.have_new_references(
        [ExtendedSWHID.from_string(swhid) for swhid in swhids]
    )
    assert result is False


def test_remover_have_new_references_missing_from_storage(
    mocker,
    sample_populated_storage,
    remover,
):
    storage = sample_populated_storage
    swhids = [
        "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
        "swh:1:snp:0000000000000000000000000000000000000022",
        "swh:1:rel:0000000000000000000000000000000000000021",
        "swh:1:rev:0000000000000000000000000000000000000018",
        "swh:1:rev:0000000000000000000000000000000000000013",
        "swh:1:dir:0000000000000000000000000000000000000017",
        "swh:1:cnt:0000000000000000000000000000000000000015",
        "swh:1:cnt:0000000000000000000000000000000000000014",
    ]
    # Patch `object_find_recent_references()` to return a SWHID
    # that does not exist in the storage. This imitates a stale
    # entry in the `object_references` table.
    mocker.patch.object(
        storage,
        "object_find_recent_references",
        wraps=lambda s, _: (
            [
                ExtendedSWHID.from_string(
                    "swh:1:rev:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                )
            ]
            if s.object_type == ExtendedObjectType.DIRECTORY
            else []
        ),
    )
    result = remover.have_new_references(
        [ExtendedSWHID.from_string(swhid) for swhid in swhids]
    )
    # As the reference actually does not exist, this means
    # no new references have actually been made since we
    # started the process.
    assert result is False


def test_remover_remove_fails_when_new_references_have_been_added(
    mocker,
    sample_populated_storage,
    remover,
):
    swhids = [
        "swh:1:cnt:0000000000000000000000000000000000000014",
    ]
    mocker.patch.object(remover, "have_new_references", return_value=True)
    remover.swhids_to_remove = [ExtendedSWHID.from_string(swhid) for swhid in swhids]
    with pytest.raises(RemoverError, match="New references"):
        remover.remove()


def test_remover_restore_recovery_bundle(
    caplog,
    mocker,
    sample_populated_storage,
    graph_client_with_only_initial_origin,
    secret_sharing_conf,
    tmp_path,
):
    from ..progressbar import no_progressbar

    bundle_path = tmp_path / "test.swh-recovery-bundle"
    mock = mocker.patch("swh.alter.operations.RecoveryBundle", autospec=True)
    instance = mock.return_value
    instance.restore.return_value = {
        "origin:add": 1,
        "origin_visit:add": 1,
        "origin_visit_status:add": 1,
    }
    restoration_storage = mocker.Mock(spec=StorageInterface)

    remover = Remover(
        storage=sample_populated_storage,
        graph_client=graph_client_with_only_initial_origin,
        restoration_storage=restoration_storage,
    )

    swhids = [
        ExtendedSWHID.from_string("swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165")
    ]
    remover.create_recovery_bundle(
        secret_sharing=SecretSharing.from_dict(secret_sharing_conf),
        requested=[swhids[0]],
        removable=Removable(removable_swhids=swhids, referencing=[]),
        recovery_bundle_path=bundle_path,
        removal_identifier="test",
    )

    with caplog.at_level(logging.INFO):
        remover.restore_recovery_bundle()
    assert "3 objects restored" in caplog.text
    assert "Something might be wrong" not in caplog.text

    instance.restore.assert_called_once_with(restoration_storage, no_progressbar)


def test_remover_restore_recovery_bundle_logs_insert_count_mismatch(
    caplog,
    mocker,
    sample_populated_storage,
    graph_client_with_only_initial_origin,
    tmp_path,
):
    mock = mocker.patch("swh.alter.operations.RecoveryBundle", autospec=True)
    instance = mock.return_value
    instance.restore.return_value = {"origin:add": 1}
    restoration_storage = mocker.Mock(spec=StorageInterface)

    remover = Remover(
        storage=sample_populated_storage,
        graph_client=graph_client_with_only_initial_origin,
        restoration_storage=restoration_storage,
    )
    # Force a path. It’ll use the mock anyway
    remover.recovery_bundle_path = tmp_path / "nonexistent.swh-recovery-bundle"

    with caplog.at_level(logging.DEBUG):
        remover.restore_recovery_bundle()

    # We force a mismatch situation. Make sure this one is unpopulated:
    assert remover.journal_objects_to_remove == {}

    assert "Something might be wrong!" in caplog.text


def test_remover_register_objects_from_bundle(
    request,
    mocker,
    remover,
    sample_recovery_bundle_path,
):
    obj_swhids: Set[str] = set()
    # We cannot use a Set as dict are not hashable
    obj_unique_keys: List[Any] = []

    def register_object(obj: BaseModel):
        if hasattr(obj, "swhid"):
            obj_swhids.add(str(obj.swhid()))
        obj_unique_keys.append(obj.unique_key())

    mocker.patch.object(remover, "register_object", side_effect=register_object)

    remover.register_objects_from_bundle(
        recovery_bundle_path=sample_recovery_bundle_path,
        object_secret_key=OBJECT_SECRET_KEY,
    )

    expected_swhids = {
        "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
        "swh:1:cnt:c932c7649c6dfa4b82327d121215116909eb3bea",
        "swh:1:cnt:33e45d56f88993aae6a0198013efa80716fd8920",
        "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
        "swh:1:dir:4b825dc642cb6eb9a060e54bf8d69288fbee4904",
        "swh:1:dir:afa0105cfcaa14fdbacee344e96659170bb1bda5",
        "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
        "swh:1:rev:a646dd94c912829659b22a1e7e143d2fa5ebde1b",
        "swh:1:rel:f7f222093a18ec60d781070abec4a630c850b837",
        "swh:1:rel:db81a26783a3f4a9db07b4759ffc37621f159bb2",
        "swh:1:snp:9b922e6d8d5b803c1582aabe5525b7b91150788e",
        "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
        "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645",
        "swh:1:ori:9147ab9c9287940d4fdbe95d8780664d7ad2dfc0",
    }
    if "version-1" not in request.keywords:
        expected_swhids |= {
            "swh:1:emd:101d70c3574c1e4b730d7ba8e83a4bdadc8691cb",
            "swh:1:emd:43dad4d96edf2fb4f77f0dbf72113b8fe8b5b664",
            "swh:1:emd:9cafd9348f3a7729c2ef0b9b149ba421589427f0",
            "swh:1:emd:ef3b0865c7a05f79772a3189ddfc8515ec3e1844",
        }
    assert obj_swhids == expected_swhids
    expected_unique_keys = [
        bytes.fromhex("3e21cc4942a4234c9e5edd8a9cacd1670fe59f13"),
        bytes.fromhex("34973274ccef6ab4dfaaf86599792fa9c3fe4689"),
        {
            "sha1": bytes.fromhex("43e45d56f88993aae6a0198013efa80716fd8920"),
            "sha1_git": bytes.fromhex("33e45d56f88993aae6a0198013efa80716fd8920"),
            "sha256": bytes.fromhex(
                "7bbd052ab054ef222c1c87be60cd191addedd24cc882d1f5f7f7be61dc61bb3a"
            ),
            "blake2s256": bytes.fromhex(
                "ade18b1adecb33f891ca36664da676e12c772cc193778aac9a137b8dc5834b9b"
            ),
        },
        bytes.fromhex("4b825dc642cb6eb9a060e54bf8d69288fbee4904"),
        bytes.fromhex("5256e856a0a0898966d6ba14feb4388b8b82d302"),
        bytes.fromhex("afa0105cfcaa14fdbacee344e96659170bb1bda5"),
        bytes.fromhex("01a7114f36fddd5ef2511b2cadda237a68adbb12"),
        bytes.fromhex("a646dd94c912829659b22a1e7e143d2fa5ebde1b"),
        bytes.fromhex("db81a26783a3f4a9db07b4759ffc37621f159bb2"),
        bytes.fromhex("f7f222093a18ec60d781070abec4a630c850b837"),
        bytes.fromhex("9b922e6d8d5b803c1582aabe5525b7b91150788e"),
        bytes.fromhex("db99fda25b43dc5cd90625ee4b0744751799c917"),
    ]
    if "version-1" not in request.keywords:
        expected_unique_keys.extend(
            [
                # RawExtrinsicMetadata
                bytes.fromhex("101d70c3574c1e4b730d7ba8e83a4bdadc8691cb"),
                bytes.fromhex("ef3b0865c7a05f79772a3189ddfc8515ec3e1844"),
                bytes.fromhex("43dad4d96edf2fb4f77f0dbf72113b8fe8b5b664"),
                bytes.fromhex("9cafd9348f3a7729c2ef0b9b149ba421589427f0"),
                # ExtID
                bytes.fromhex("486e20ccedc221075b12abbb607a888875db41f6"),
                bytes.fromhex("fa730cf0bb415e1e921e430984bdcddd9c8eea4a"),
            ]
        )
    expected_unique_keys.extend(
        [
            {"url": "https://github.com/user1/repo1"},
            {
                "origin": "https://github.com/user1/repo1",
                "date": "2015-01-01 23:00:00+00:00",
            },
            {
                "origin": "https://github.com/user1/repo1",
                "date": "2017-01-01 23:00:00+00:00",
            },
            {
                "origin": "https://github.com/user1/repo1",
                "visit": "1",
                "date": "2015-01-01 23:00:00+00:00",
            },
            {
                "origin": "https://github.com/user1/repo1",
                "visit": "2",
                "date": "2017-01-01 23:00:00+00:00",
            },
            {"url": "https://github.com/user2/repo1"},
            {
                "origin": "https://github.com/user2/repo1",
                "date": "2015-01-01 23:00:00+00:00",
            },
            {
                "origin": "https://github.com/user2/repo1",
                "visit": "1",
                "date": "2015-01-01 23:00:00+00:00",
            },
        ]
    )
    assert obj_unique_keys == expected_unique_keys


def test_remover_handle_removal_notification_with_removal(
    mocker,
    tmp_path,
    sample_populated_storage_with_matching_hash,
    empty_graph_client,
    populated_masking_admin,
    example_removal_notification_with_matching_hash,
    secret_sharing_conf,
):
    storage = sample_populated_storage_with_matching_hash
    skipped_content_swhid = ExtendedSWHID.from_string(
        # SkippedContent
        "swh:1:cnt:33e45d56f88993aae6a0198013efa80716fd8920"
    )
    removable_swhids = example_removal_notification_with_matching_hash.removed_objects
    removable_swhids.remove(skipped_content_swhid)

    referencing = [
        ExtendedSWHID.from_string(s)
        for s in (
            "swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa",
            "swh:1:dir:8505808532953da7d2581741f01b29c04b1cb9ab",
        )
    ]

    masking_admin = populated_masking_admin
    notification_removal_identifier = (
        example_removal_notification_with_matching_hash.removal_identifier
    )
    recovery_bundle_path = tmp_path / "test.swh-bundle"
    remover = Remover(
        storage=storage,
        restoration_storage=storage,
        graph_client=empty_graph_client,
        masking_admin=masking_admin,
    )
    mock_get_removable = mocker.patch.object(
        remover,
        "get_removable",
        return_value=Removable(
            removable_swhids=removable_swhids, referencing=referencing
        ),
    )
    mocker.patch.object(remover, "have_new_references", return_value=False)

    remover.handle_removal_notification_with_removal(
        notification_removal_identifier=notification_removal_identifier,
        secret_sharing=SecretSharing.from_dict(secret_sharing_conf),
        recovery_bundle_path=recovery_bundle_path,
        ignore_requested=[
            Origin(url="https://github.com/user1/repo1"),
            ExtendedSWHID.from_string(
                "swh:1:ori:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
            ),
        ],
    )

    mock_get_removable.assert_called_once_with(
        [
            ExtendedSWHID.from_string(
                "swh:1:ori:9147ab9c9287940d4fdbe95d8780664d7ad2dfc0"
            )
        ]
    )

    with masking_admin.conn:
        masking_request = masking_admin.find_request(
            f"{MASKING_REQUEST_IDENTIFIER_PREFIX}{notification_removal_identifier}"
        )
        masking_states = masking_admin.get_states_for_request(masking_request.id)
        assert masking_states == {
            swhid: MaskedState.VISIBLE for swhid in removable_swhids
        } | {skipped_content_swhid: MaskedState.DECISION_PENDING}
        history = masking_admin.get_history(masking_request.id)
        assert list(sorted(history, key=operator.attrgetter("date")))[
            -1
        ].message == textwrap.dedent(
            """\
            Made 17 objects visible again after removal.

            Objects in state “decision pending” that were not removed:
            - swh:1:cnt:33e45d56f88993aae6a0198013efa80716fd8920"""
        )


def test_remover_handle_removal_notification_with_removal_masked_request_not_found(
    tmp_path,
    sample_populated_storage_with_matching_hash,
    empty_graph_client,
    populated_masking_admin,
    example_removal_notification_with_matching_hash,
    secret_sharing_conf,
):
    storage = sample_populated_storage_with_matching_hash
    remover = Remover(
        storage=storage,
        restoration_storage=storage,
        graph_client=empty_graph_client,
        masking_admin=populated_masking_admin,
    )
    recovery_bundle_path = tmp_path / "test.swh-bundle"
    with pytest.raises(MaskingRequestNotFound, match="NON_EXISTENT"):
        remover.handle_removal_notification_with_removal(
            notification_removal_identifier="NON_EXISTENT",
            secret_sharing=SecretSharing.from_dict(secret_sharing_conf),
            recovery_bundle_path=recovery_bundle_path,
        )


@pytest.mark.parametrize(
    "masked_status, expected_message",
    [
        (MaskedState.RESTRICTED, "Made 18 objects permanently restricted."),
        (
            MaskedState.DECISION_PENDING,
            "Made 18 objects restricted until a decision is made.",
        ),
        (MaskedState.VISIBLE, "Made 18 objects visible again."),
    ],
)
def test_handle_removal_notification_by_changing_masked_status(
    masked_status,
    expected_message,
    sample_populated_storage_with_matching_hash,
    empty_graph_client,
    populated_masking_admin,
    example_removal_notification_with_matching_hash,
):
    storage = sample_populated_storage_with_matching_hash
    masking_admin = populated_masking_admin
    remover = Remover(
        storage=storage,
        restoration_storage=storage,
        graph_client=empty_graph_client,
        masking_admin=masking_admin,
    )
    notification_removal_identifier = (
        example_removal_notification_with_matching_hash.removal_identifier
    )
    remover.handle_removal_notification_by_changing_masked_status(
        notification_removal_identifier=notification_removal_identifier,
        masked_state=masked_status,
    )
    with masking_admin.conn:
        masking_request = masking_admin.find_request(
            f"{MASKING_REQUEST_IDENTIFIER_PREFIX}{notification_removal_identifier}"
        )
        history = masking_admin.get_history(masking_request.id)
        assert (
            list(sorted(history, key=operator.attrgetter("date")))[-1].message
            == expected_message
        )
        assert all(
            status == masked_status
            for status in masking_admin.get_states_for_request(
                masking_request.id
            ).values()
        )


def test_handle_removal_notification_by_changing_masked_status_masked_request_not_found(
    sample_populated_storage_with_matching_hash,
    empty_graph_client,
    populated_masking_admin,
):
    storage = sample_populated_storage_with_matching_hash
    remover = Remover(
        storage=storage,
        restoration_storage=storage,
        graph_client=empty_graph_client,
        masking_admin=populated_masking_admin,
    )
    with pytest.raises(MaskingRequestNotFound, match="NON_EXISTENT"):
        remover.handle_removal_notification_by_changing_masked_status(
            notification_removal_identifier="NON_EXISTENT",
            masked_state=MaskedState.RESTRICTED,
        )



# Copyright (C) 2023-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

import swh.graph.example_dataset as graph_dataset
from swh.graph.example_dataset import INITIAL_ORIGIN
from swh.model.swhids import ExtendedSWHID

from ..inventory import InventorySubgraph
from ..removable import mark_removable
from .conftest import write_dot_if_requested


@pytest.fixture
def inventory_from_initial_origin():
    g = InventorySubgraph()
    v_ori = g.add_swhid(INITIAL_ORIGIN.swhid())
    v_snp = g.add_swhid("swh:1:snp:0000000000000000000000000000000000000020")
    v_rel = g.add_swhid("swh:1:rel:0000000000000000000000000000000000000010")
    v_rev_09 = g.add_swhid("swh:1:rev:0000000000000000000000000000000000000009")
    v_rev_03 = g.add_swhid("swh:1:rev:0000000000000000000000000000000000000003")
    v_dir_08 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000008")
    v_dir_06 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000006")
    v_dir_02 = g.add_swhid("swh:1:dir:0000000000000000000000000000000000000002")
    v_cnt_07 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000007")
    v_cnt_05 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000005")
    v_cnt_04 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000004")
    v_cnt_01 = g.add_swhid("swh:1:cnt:0000000000000000000000000000000000000001")
    g.add_edge(v_ori, v_snp)
    g.add_edge(v_snp, v_rel)
    g.add_edge(v_snp, v_rev_09)
    g.add_edge(v_rel, v_rev_09)
    g.add_edge(v_rev_09, v_rev_03)
    g.add_edge(v_rev_09, v_dir_08)
    g.add_edge(v_rev_03, v_dir_02)
    g.add_edge(v_dir_08, v_dir_06)
    g.add_edge(v_dir_08, v_cnt_07)
    g.add_edge(v_dir_08, v_cnt_01)
    g.add_edge(v_dir_06, v_cnt_05)
    g.add_edge(v_dir_06, v_cnt_04)
    g.add_edge(v_dir_02, v_cnt_01)
    write_dot_if_requested(g, "inventory_from_initial_origin.dot")
    return g


@pytest.fixture
def storage_with_no_new_references_since_export(mocker, sample_populated_storage):
    # Simulate an empty `object_references` table
    mocker.patch.object(
        sample_populated_storage,
        "object_find_recent_references",
        create=True,
        return_value=[],
    )
    return sample_populated_storage


@pytest.mark.parametrize(
    "graph_client_fixture, storage_fixture",
    [
        (
            "graph_client_with_both_origins",
            "storage_with_no_new_references_since_export",
        ),
        (
            "graph_client_with_only_initial_origin",
            "sample_populated_storage",
        ),
        (
            "graph_client_with_both_origins",
            "sample_populated_storage",
        ),
    ],
)
def test_mark_removable_on_initial_origin(
    request, storage_fixture, graph_client_fixture, inventory_from_initial_origin
):
    storage = request.getfixturevalue(storage_fixture)
    graph_client = request.getfixturevalue(graph_client_fixture)
    subgraph = mark_removable(storage, graph_client, inventory_from_initial_origin)
    # We are trying to remove the initial origin. As everything is used by the
    # fork, the only things that can be removed in the end are the corresponding
    # origin and snapshot.
    assert {str(swhid) for swhid in subgraph.removable_swhids()} == {
        "swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054",
        "swh:1:snp:0000000000000000000000000000000000000020",
    }


@pytest.mark.parametrize(
    "graph_client_fixture, storage_fixture",
    [
        (
            "graph_client_with_both_origins",
            "storage_with_no_new_references_since_export",
        ),
        (
            "graph_client_with_only_initial_origin",
            "sample_populated_storage",
        ),
        (
            "graph_client_with_both_origins",
            "sample_populated_storage",
        ),
        (
            "graph_client_with_submodule",
            # This one is not ideal as it does not contain our origin with
            # submodule. This should not matter for testing that we have
            # the right behavior in presence of submodules, though.
            "storage_with_no_new_references_since_export",
        ),
    ],
)
def test_mark_removable_on_forked_origin(
    request, storage_fixture, graph_client_fixture, inventory_from_forked_origin
):
    storage = request.getfixturevalue(storage_fixture)
    graph_client = request.getfixturevalue(graph_client_fixture)
    subgraph = mark_removable(storage, graph_client, inventory_from_forked_origin)
    # We are trying to remove the forked origin. Therefore objects coming from the
    # initial origin are unremovable. So we are left removing all objects specific
    # to the development that happened in the forked origin.
    assert {str(swhid) for swhid in subgraph.removable_swhids()} == {
        "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
        "swh:1:snp:0000000000000000000000000000000000000022",
        "swh:1:rel:0000000000000000000000000000000000000021",
        "swh:1:rev:0000000000000000000000000000000000000018",
        "swh:1:rev:0000000000000000000000000000000000000013",
        "swh:1:dir:0000000000000000000000000000000000000017",
        "swh:1:dir:0000000000000000000000000000000000000016",
        "swh:1:dir:0000000000000000000000000000000000000012",
        "swh:1:cnt:0000000000000000000000000000000000000015",
        "swh:1:cnt:0000000000000000000000000000000000000014",
        "swh:1:cnt:0000000000000000000000000000000000000011",
    }


def test_mark_removable_on_initial_origin_with_forked_origin_removed_and_oudated_graph(
    storage_with_forked_origin_removed,
    graph_client_with_both_origins,
    inventory_from_initial_origin,
):
    # Test the case of an outdated graph
    subgraph = mark_removable(
        storage_with_forked_origin_removed,
        graph_client_with_both_origins,
        inventory_from_initial_origin,
    )
    assert {str(swhid) for swhid in subgraph.removable_swhids()} == {
        "swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054",
        "swh:1:snp:0000000000000000000000000000000000000020",
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000009",
        "swh:1:rev:0000000000000000000000000000000000000003",
        "swh:1:dir:0000000000000000000000000000000000000008",
        "swh:1:dir:0000000000000000000000000000000000000006",
        "swh:1:dir:0000000000000000000000000000000000000002",
        "swh:1:cnt:0000000000000000000000000000000000000007",
        "swh:1:cnt:0000000000000000000000000000000000000005",
        "swh:1:cnt:0000000000000000000000000000000000000004",
        "swh:1:cnt:0000000000000000000000000000000000000001",
    }


def test_mark_removable_on_stale_object_references_table(
    swh_storage,
    empty_graph_client,
    inventory_from_initial_origin,
):
    # Ensure we have the right revision pointing to the right directory
    directory = graph_dataset.DIRECTORIES[0]
    revision = graph_dataset.REVISIONS[0]
    assert directory.id == revision.directory

    swh_storage.revision_add([revision])
    swh_storage.directory_add([directory])
    result = swh_storage.flush()
    assert result == {"revision:add": 1, "directory:add": 1, "object_reference:add": 2}

    result = swh_storage.object_delete([revision.swhid().to_extended()])
    assert result["revision:delete"] == 1

    # Now the `object_references` table is outdated because it
    # contains a stale entry with “revision → directory”.
    # What if we try to remove the directory now?
    inventory_subgraph = InventorySubgraph()
    inventory_subgraph.add_swhid(str(directory.swhid()))
    subgraph = mark_removable(
        swh_storage,
        empty_graph_client,
        inventory_subgraph,
    )
    assert {str(swhid) for swhid in subgraph.removable_swhids()} == {
        str(directory.swhid())
    }


def test_mark_removable_with_known_missing(
    storage_with_forked_origin_removed,
    graph_client_with_only_initial_origin,
    inventory_from_initial_origin,
):
    known_missing = {
        ExtendedSWHID.from_string("swh:1:dir:0000000000000000000000000000000000000006"),
    }
    # Test the case of an outdated graph
    subgraph = mark_removable(
        storage_with_forked_origin_removed,
        graph_client_with_only_initial_origin,
        inventory_from_initial_origin,
        known_missing=known_missing,
    )
    assert {str(swhid) for swhid in subgraph.removable_swhids()} == {
        "swh:1:ori:83404f995118bd25774f4ac14422a8f175e7a054",
        "swh:1:snp:0000000000000000000000000000000000000020",
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000009",
        "swh:1:rev:0000000000000000000000000000000000000003",
        "swh:1:dir:0000000000000000000000000000000000000008",
        "swh:1:dir:0000000000000000000000000000000000000002",
        "swh:1:cnt:0000000000000000000000000000000000000007",
        "swh:1:cnt:0000000000000000000000000000000000000005",
        "swh:1:cnt:0000000000000000000000000000000000000004",
        "swh:1:cnt:0000000000000000000000000000000000000001",
    }
    # Normally we would have removed everything, but objects
    # known missing are not considered removable, therefore
    # they won’t be in the recovery bundle, and thus are
    # referenced by the set of removable objects.
    assert set(subgraph.referenced_swhids()) == known_missing



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import partial
import itertools
import operator
import random
from typing import Iterable, List

import pytest

from swh.model.swhids import ExtendedObjectType, ExtendedSWHID

from ..utils import iter_swhids_grouped_by_type


@pytest.fixture
def randomized_swhid_list() -> List[ExtendedSWHID]:
    swhids = [
        # Content
        "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
        "swh:1:cnt:c932c7649c6dfa4b82327d121215116909eb3bea",
        # SkippedContent
        "swh:1:cnt:33e45d56f88993aae6a0198013efa80716fd8920",
        # Directory
        "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
        "swh:1:dir:4b825dc642cb6eb9a060e54bf8d69288fbee4904",
        "swh:1:dir:afa0105cfcaa14fdbacee344e96659170bb1bda5",
        # Revision
        "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
        "swh:1:rev:a646dd94c912829659b22a1e7e143d2fa5ebde1b",
        # Release
        "swh:1:rel:f7f222093a18ec60d781070abec4a630c850b837",
        "swh:1:rel:db81a26783a3f4a9db07b4759ffc37621f159bb2",
        # Snapshot
        "swh:1:snp:9b922e6d8d5b803c1582aabe5525b7b91150788e",
        "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
        # Origin
        "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645",
        "swh:1:ori:9147ab9c9287940d4fdbe95d8780664d7ad2dfc0",
    ]
    random.shuffle(swhids)
    return [ExtendedSWHID.from_string(swhid) for swhid in swhids]


def test_iter_swhids_grouped_by_type(
    randomized_swhid_list: List[ExtendedSWHID],
) -> None:
    def handle_swhids(
        object_type: ExtendedObjectType, swhids: List[ExtendedSWHID]
    ) -> Iterable[ExtendedSWHID]:
        assert all(swhid.object_type == object_type for swhid in swhids)
        yield from swhids

    def chunker(swhids):
        yield list(swhids)

    handlers = {
        object_type: partial(handle_swhids, object_type)
        for object_type in [
            ExtendedObjectType.CONTENT,
            ExtendedObjectType.DIRECTORY,
            ExtendedObjectType.REVISION,
            ExtendedObjectType.RELEASE,
            ExtendedObjectType.SNAPSHOT,
            ExtendedObjectType.ORIGIN,
        ]
    }
    grouped_results = list(
        set(str(swhid) for swhid in swhids)
        for _, swhids in itertools.groupby(
            iter_swhids_grouped_by_type(
                randomized_swhid_list, handlers=handlers, chunker=chunker
            ),
            key=operator.attrgetter("object_type"),
        )
    )
    assert grouped_results == [
        # Content & SkippedContent
        {
            "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
            "swh:1:cnt:c932c7649c6dfa4b82327d121215116909eb3bea",
            "swh:1:cnt:33e45d56f88993aae6a0198013efa80716fd8920",
        },
        # Directory
        {
            "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
            "swh:1:dir:4b825dc642cb6eb9a060e54bf8d69288fbee4904",
            "swh:1:dir:afa0105cfcaa14fdbacee344e96659170bb1bda5",
        },
        # Revision
        {
            "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
            "swh:1:rev:a646dd94c912829659b22a1e7e143d2fa5ebde1b",
        },
        # Release
        {
            "swh:1:rel:f7f222093a18ec60d781070abec4a630c850b837",
            "swh:1:rel:db81a26783a3f4a9db07b4759ffc37621f159bb2",
        },
        # Snapshot
        {
            "swh:1:snp:9b922e6d8d5b803c1582aabe5525b7b91150788e",
            "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
        },
        # Origin
        {
            "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645",
            "swh:1:ori:9147ab9c9287940d4fdbe95d8780664d7ad2dfc0",
        },
    ]


def test_iter_swhids_grouped_by_type_no_chunker(
    randomized_swhid_list: List[ExtendedSWHID],
) -> None:
    def handle_swhids(
        object_type: ExtendedObjectType, swhids: Iterable[ExtendedSWHID]
    ) -> Iterable[ExtendedSWHID]:
        yield from swhids

    handlers = {
        object_type: partial(handle_swhids, object_type)
        for object_type in [
            ExtendedObjectType.CONTENT,
            ExtendedObjectType.DIRECTORY,
            ExtendedObjectType.REVISION,
            ExtendedObjectType.RELEASE,
            ExtendedObjectType.SNAPSHOT,
            ExtendedObjectType.ORIGIN,
        ]
    }
    grouped_results = list(
        set(str(swhid) for swhid in swhids)
        for _, swhids in itertools.groupby(
            iter_swhids_grouped_by_type(randomized_swhid_list, handlers=handlers),
            key=operator.attrgetter("object_type"),
        )
    )
    assert grouped_results == [
        # Content & SkippedContent
        {
            "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
            "swh:1:cnt:c932c7649c6dfa4b82327d121215116909eb3bea",
            "swh:1:cnt:33e45d56f88993aae6a0198013efa80716fd8920",
        },
        # Directory
        {
            "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
            "swh:1:dir:4b825dc642cb6eb9a060e54bf8d69288fbee4904",
            "swh:1:dir:afa0105cfcaa14fdbacee344e96659170bb1bda5",
        },
        # Revision
        {
            "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
            "swh:1:rev:a646dd94c912829659b22a1e7e143d2fa5ebde1b",
        },
        # Release
        {
            "swh:1:rel:f7f222093a18ec60d781070abec4a630c850b837",
            "swh:1:rel:db81a26783a3f4a9db07b4759ffc37621f159bb2",
        },
        # Snapshot
        {
            "swh:1:snp:9b922e6d8d5b803c1582aabe5525b7b91150788e",
            "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
        },
        # Origin
        {
            "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645",
            "swh:1:ori:9147ab9c9287940d4fdbe95d8780664d7ad2dfc0",
        },
    ]



# Copyright (C) 2023-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
from datetime import datetime, timedelta
from functools import partial, reduce
import itertools
import logging
import operator
import re
import statistics
import time
from typing import (
    Any,
    Dict,
    FrozenSet,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    TextIO,
    Tuple,
    cast,
)

import humanize
from tabulate import tabulate
import yaml

from swh.core.utils import grouper
from swh.graph.http_client import RemoteGraphClient
from swh.journal.writer.kafka import KafkaJournalWriter
from swh.model.model import BaseModel, Content, KeyType, Origin
from swh.model.swhids import CoreSWHID, ExtendedObjectType, ExtendedSWHID
from swh.model.swhids import ObjectType as CoreSWHIDObjectType
from swh.objstorage.exc import Error as ObjstorageError
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.interface import (
    CompositeObjId,
    ObjStorageInterface,
    objid_from_dict,
)
from swh.search.interface import SearchInterface
from swh.storage.interface import ObjectDeletionInterface, StorageInterface
from swh.storage.proxies.masking.db import MaskedState, MaskingAdmin, MaskingRequest

from .inventory import RootsNotFound, get_raw_extrinsic_metadata, make_inventory
from .mirror_notification_watcher import MASKING_REQUEST_IDENTIFIER_PREFIX
from .progressbar import ProgressBar, ProgressBarInit, no_progressbar
from .recovery_bundle import (
    AgeSecretKey,
    RecoveryBundle,
    RecoveryBundleCreator,
    SecretSharing,
    generate_age_keypair,
)
from .removable import mark_removable
from .utils import get_filtered_objects

logger = logging.getLogger(__name__)


OBJSTORAGE_DELETE_MAX_ATTEMPTS = 3


class RemoverError(Exception):
    pass


class MaskingRequestNotFound(Exception):
    def __init__(self, masking_request_slug: str):
        self.masking_request_slug = masking_request_slug

    def __str__(self) -> str:
        return f"Masking request “{self.masking_request_slug}” not found."


def _secho(msg, **kwargs):
    """Log at info level, passing kwargs as styles for click.secho()"""
    logger.info(msg, extra={"style": kwargs})


def format_duration(seconds: float) -> str:
    return humanize.precisedelta(
        timedelta(seconds=seconds), minimum_unit="milliseconds"
    )


class Removable(NamedTuple):
    """Aggregates information returned by :py:meth:`Remover.get_removable` to
    be then used with :py:meth:`Remover.create_recovery_bundle`."""

    removable_swhids: List[ExtendedSWHID]
    referencing: List[ExtendedSWHID]

    def print_plan(self) -> None:
        ordering = {
            t: i
            for i, t in enumerate(
                (
                    ExtendedObjectType.ORIGIN,
                    ExtendedObjectType.SNAPSHOT,
                    ExtendedObjectType.RELEASE,
                    ExtendedObjectType.REVISION,
                    ExtendedObjectType.DIRECTORY,
                    ExtendedObjectType.CONTENT,
                    ExtendedObjectType.RAW_EXTRINSIC_METADATA,
                )
            )
        }
        sorted_swhids = sorted(
            self.removable_swhids, key=lambda swhid: ordering[swhid.object_type]
        )
        _secho("Removal plan:")
        for object_type, grouped_swhids in itertools.groupby(
            sorted_swhids, key=operator.attrgetter("object_type")
        ):
            _secho(f"- {object_type.name.capitalize()}: {len(list(grouped_swhids))}")
        _secho(
            "- … and more objects that are not addresseable by a SWHID "
            "(OriginVisit, OriginVisitStatus, ExtID)."
        )


STORAGE_OBJECT_DELETE_CHUNK_SIZE = 200


class Remover:
    """Helper class used to perform a removal."""

    def __init__(
        self,
        /,
        storage: StorageInterface,
        graph_client: Optional[RemoteGraphClient] = None,
        known_missing: Optional[Set[ExtendedSWHID]] = None,
        restoration_storage: Optional[StorageInterface] = None,
        removal_searches: Optional[Dict[str, SearchInterface]] = None,
        removal_storages: Optional[Dict[str, ObjectDeletionInterface]] = None,
        removal_objstorages: Optional[Dict[str, ObjStorageInterface]] = None,
        removal_journals: Optional[Dict[str, KafkaJournalWriter]] = None,
        masking_admin: Optional[MaskingAdmin] = None,
        progressbar: Optional[ProgressBarInit] = None,
    ):
        self.storage = storage
        self.graph_client = graph_client
        self.known_missing = known_missing or set()
        self.restoration_storage = restoration_storage
        self.removal_searches = removal_searches if removal_searches else {}
        self.removal_storages = removal_storages if removal_storages else {}
        self.removal_objstorages = removal_objstorages if removal_objstorages else {}
        self.removal_journals = removal_journals if removal_journals else {}
        self.masking_admin = masking_admin
        self.recovery_bundle_path: Optional[str] = None
        self.object_secret_key: Optional[AgeSecretKey] = None
        self.swhids_to_remove: List[ExtendedSWHID] = []
        self.objids_to_remove: List[CompositeObjId] = []
        self.origin_urls_to_remove: List[str] = []
        self.journal_objects_to_remove: Dict[str, List[KeyType]] = (
            collections.defaultdict(list)
        )
        self.progressbar: ProgressBarInit = (
            progressbar if progressbar is not None else no_progressbar
        )

    def get_removable(
        self,
        swhids: List[ExtendedSWHID],
        *,
        output_inventory_subgraph: Optional[TextIO] = None,
        output_removable_subgraph: Optional[TextIO] = None,
        output_pruned_removable_subgraph: Optional[TextIO] = None,
    ) -> Removable:
        _secho("Removing the following origins:")
        for swhid in swhids:
            _secho(f" - {swhid}")
        _secho("Finding removable objects…", fg="cyan")
        assert self.graph_client is not None
        inventory_subgraph = make_inventory(
            self.storage,
            self.graph_client,
            swhids,
            known_missing=self.known_missing,
            progressbar=self.progressbar,
        )
        if output_inventory_subgraph:
            inventory_subgraph.write_dot(output_inventory_subgraph)
            output_inventory_subgraph.close()
        removable_subgraph = mark_removable(
            self.storage,
            self.graph_client,
            inventory_subgraph,
            self.known_missing,
            self.progressbar,
        )
        if output_removable_subgraph:
            removable_subgraph.write_dot(output_removable_subgraph)
            output_removable_subgraph.close()
        referencing = list(removable_subgraph.referenced_swhids())
        removable_swhids = list(removable_subgraph.removable_swhids())
        removable_swhids.extend(
            get_raw_extrinsic_metadata(
                self.storage, removable_swhids, progressbar=self.progressbar
            )
        )
        if output_pruned_removable_subgraph:
            removable_subgraph.delete_unremovable()
            removable_subgraph.write_dot(output_pruned_removable_subgraph)
            output_pruned_removable_subgraph.close()
        return Removable(
            removable_swhids=removable_swhids,
            referencing=referencing,
        )

    def register_object(self, obj: BaseModel) -> None:
        # Register for removal from storage
        if hasattr(obj, "swhid"):
            # StorageInterface.ObjectDeletionInterface.remove uses SWHIDs
            # for reference. We hope it will handle objects without SWHIDs
            # (origin_visit, origin_visit_status) directly.
            obj_swhid = obj.swhid()
            if obj_swhid is not None:
                swhid = (
                    obj_swhid.to_extended()
                    if isinstance(obj_swhid, CoreSWHID)
                    else obj_swhid
                )
                self.swhids_to_remove.append(swhid)
                if swhid.object_type == ExtendedObjectType.CONTENT:
                    content = cast(Content, obj)
                    self.objids_to_remove.append(objid_from_dict(content.to_dict()))
        # Register for removal from the journal
        self.journal_objects_to_remove[str(obj.object_type)].append(obj.unique_key())
        # Register for removal from search
        if isinstance(obj, Origin):
            self.origin_urls_to_remove.append(obj.url)

    def register_objects_from_bundle(
        self, recovery_bundle_path: str, object_secret_key: AgeSecretKey
    ) -> RecoveryBundle:
        assert self.recovery_bundle_path is None
        assert self.object_secret_key is None

        def key_provider(_):
            return object_secret_key

        bundle = RecoveryBundle(recovery_bundle_path, key_provider)
        _secho(
            f"Resuming removal from bundle “{bundle.removal_identifier}”…",
            fg="cyan",
            bold=True,
        )
        self.recovery_bundle_path = recovery_bundle_path
        self.object_secret_key = object_secret_key

        iterchain = [
            bundle.contents(),
            bundle.skipped_contents(),
            bundle.directories(),
            bundle.revisions(),
            bundle.releases(),
            bundle.snapshots(),
        ]
        if bundle.version >= 2:
            iterchain.extend(
                [
                    bundle.raw_extrinsic_metadata(),
                    bundle.extids(),
                ]
            )
        bar: ProgressBar[int]
        with self.progressbar(
            length=len(bundle.swhids), label="Loading objects…"
        ) as bar:
            for obj in itertools.chain(*iterchain):
                self.register_object(obj)
                bar.update(n_steps=1)
            for origin in bundle.origins():
                self.register_object(origin)
                for obj in itertools.chain(
                    bundle.origin_visits(origin), bundle.origin_visit_statuses(origin)
                ):
                    self.register_object(obj)
                bar.update(n_steps=1)
        return bundle

    def create_recovery_bundle(
        self,
        /,
        secret_sharing: SecretSharing,
        requested: List[Origin | ExtendedSWHID],
        removable: Removable,
        recovery_bundle_path: str,
        removal_identifier: str,
        reason: Optional[str] = None,
        expire: Optional[datetime] = None,
        allow_empty_content_objects: bool = False,
    ) -> AgeSecretKey:
        object_public_key, self.object_secret_key = generate_age_keypair()
        decryption_key_shares = secret_sharing.generate_encrypted_shares(
            removal_identifier, self.object_secret_key
        )
        _secho("Creating recovery bundle…", fg="cyan")
        with RecoveryBundleCreator(
            path=recovery_bundle_path,
            storage=self.storage,
            removal_identifier=removal_identifier,
            requested=requested,
            referencing=removable.referencing,
            object_public_key=object_public_key,
            decryption_key_shares=decryption_key_shares,
            registration_callback=self.register_object,
            allow_empty_content_objects=allow_empty_content_objects,
        ) as creator:
            if reason is not None:
                creator.set_reason(reason)
            if expire is not None:
                try:
                    creator.set_expire(expire)
                except ValueError as ex:
                    raise RemoverError(f"Unable to set expiration date: {str(ex)}")
            creator.backup_swhids(
                removable.removable_swhids, progressbar=self.progressbar
            )
        self.recovery_bundle_path = recovery_bundle_path
        _secho("Recovery bundle created.", fg="green")
        return self.object_secret_key

    def restore_recovery_bundle(self) -> None:
        assert self.restoration_storage
        assert self.recovery_bundle_path

        def key_provider(_):
            assert self.object_secret_key
            return self.object_secret_key

        bundle = RecoveryBundle(self.recovery_bundle_path, key_provider)
        result = bundle.restore(self.restoration_storage, self.progressbar)
        # We care about the number of objects, not the byte count
        result.pop("content:add:bytes", None)
        # We don’t care about new object_references either
        result.pop("object_reference:add", None)
        total = sum(result.values())
        _secho(f"{total} objects restored.", fg="green")
        count_from_journal_objects = sum(
            len(objects) for objects in self.journal_objects_to_remove.values()
        )
        if count_from_journal_objects != total:
            _secho(
                f"{count_from_journal_objects} objects should have "
                "been restored. Something might be wrong!",
                fg="red",
                bold=True,
            )

    def remove(self, progressbar=None) -> None:
        for name, removal_search in self.removal_searches.items():
            self.remove_from_search(name, removal_search)
        for name, removal_storage in self.removal_storages.items():
            self.remove_from_storage(name, removal_storage)
        for name, journal_writer in self.removal_journals.items():
            self.remove_from_journal(name, journal_writer)
        if len(self.removal_objstorages) > 0:
            self.remove_from_objstorages()
        if self.have_new_references(self.swhids_to_remove):
            raise RemoverError(
                "New references have been added to removed objects. "
                "This invalidates the initial set of candidates for removal."
            )

    def remove_from_storage(
        self, name: str, removal_storage: ObjectDeletionInterface
    ) -> None:
        results: collections.Counter[str] = collections.Counter()
        bar: ProgressBar[int]
        with self.progressbar(
            length=len(self.swhids_to_remove),
            label=f"Removing objects from storage “{name}”…",
        ) as bar:
            for chunk_it in grouper(
                self.swhids_to_remove, STORAGE_OBJECT_DELETE_CHUNK_SIZE
            ):
                chunk_swhids = list(chunk_it)
                # Remove objects addressable by a SWHID
                results += removal_storage.object_delete(chunk_swhids)
                # Remove ExtIDs (addressable by their targets)
                chunk_core_swhids = [
                    CoreSWHID(
                        object_type=CoreSWHIDObjectType[
                            extended_swhid.object_type.name
                        ],
                        object_id=extended_swhid.object_id,
                    )
                    for extended_swhid in chunk_swhids
                    if hasattr(CoreSWHIDObjectType, extended_swhid.object_type.name)
                ]
                results += removal_storage.extid_delete_for_target(chunk_core_swhids)
                bar.update(n_steps=len(chunk_swhids))
        _secho(
            f"{results.total()} objects removed from storage “{name}”.",
            fg="green",
        )

    def remove_from_journal(
        self, name: str, journal_writer: KafkaJournalWriter
    ) -> None:
        bar: ProgressBar[Tuple[str, List[KeyType]]]
        with self.progressbar(
            self.journal_objects_to_remove.items(),
            label=f"Removing objects from journal “{name}”…",
        ) as bar:
            for object_type, keys in bar:
                journal_writer.delete(object_type, keys)
        journal_writer.flush()
        _secho(f"Objects removed from journal “{name}”.", fg="green")

    def remove_from_search(self, name: str, search: SearchInterface) -> None:
        count = 0
        with self.progressbar(
            self.origin_urls_to_remove, label=f"Removing origins from search “{name}”…"
        ) as bar:
            for origin_url in bar:
                deleted = search.origin_delete(origin_url)
                count += 1 if deleted else 0
            search.flush()
        _secho(f"{count} origins removed from search “{name}”.", fg="green")

    def remove_from_objstorages(self):
        results = []
        for name, removal_objstorage in self.removal_objstorages.items():
            results.append(self.remove_from_objstorage(name, removal_objstorage))
        not_found = reduce(set.intersection, results)
        if not_found:
            table = tabulate(
                (
                    dict(sorted(frozen_objid, key=operator.itemgetter(0)))
                    for frozen_objid in not_found
                ),
                headers="keys",
                tablefmt="github",
                disable_numparse=True,
            )
            _secho(f"Objects not found in any objstorage:\n{table}", fg="red")

    def remove_from_objstorage(
        self,
        name: str,
        objstorage: ObjStorageInterface,
    ) -> Set[FrozenSet[Tuple[str, str]]]:
        count = 0
        not_found: Set[FrozenSet[Tuple[str, str]]] = set()
        durations = []
        with self.progressbar(
            self.objids_to_remove, label=f"Removing objects from objstorage “{name}”…"
        ) as bar:
            for objid in bar:
                attempt = 1
                while True:
                    try:
                        start = time.monotonic()
                        objstorage.delete(objid)
                        durations.append(time.monotonic() - start)
                        count += 1
                        break
                    except ObjNotFoundError:
                        # hex form is nicer to read
                        objid_hex = {k: cast(bytes, v).hex() for k, v in objid.items()}
                        # convert to a frozenset of tuples as dicts are not hashable
                        not_found.add(frozenset(objid_hex.items()))
                        logger.debug(
                            "%s not found in objstorage “%s” for deletion",
                            objid_hex,
                            name,
                        )
                        break
                    except ObjstorageError as e:
                        raise e
                    except Exception as e:
                        if attempt >= OBJSTORAGE_DELETE_MAX_ATTEMPTS:
                            raise e
                        else:
                            cooldown = 5 * attempt
                            logger.warning(
                                "objstorage “%s” raised “%r” during attempt %d, "
                                "retrying in %d seconds…",
                                name,
                                e,
                                attempt,
                                cooldown,
                            )
                            time.sleep(cooldown)
                    attempt += 1
        stats = (
            (
                f" Total time: {format_duration(sum(durations))},"
                f" average: {format_duration(statistics.mean(durations))} per object"
            )
            if len(durations) > 0
            else ""
        )
        stdev = (
            f", standard deviation: {format_duration(statistics.stdev(durations))}"
            if len(durations) >= 2
            else ""
        )
        _secho(
            f"{count} objects removed from objstorage “{name}”.{stats}{stdev}",
            fg="green",
        )
        return not_found

    def have_new_references(self, removed_swhids: List[ExtendedSWHID]) -> bool:
        """Find out if any removed objects now have a new references coming from
        an object outside the set of removed objects."""

        swhids = set(removed_swhids)
        with self.progressbar(
            swhids, label="Looking for newly added references…"
        ) as bar:
            for swhid in bar:
                if swhid.object_type == ExtendedObjectType.ORIGIN:
                    continue
                recent_references = get_filtered_objects(
                    self.storage,
                    partial(self.storage.object_find_recent_references, swhid),
                    len(swhids) + 1,
                )
                if not swhids.issuperset(set(recent_references)):
                    return True
        return False

    @staticmethod
    def _get_info_from_masking_request(
        masking_request: MaskingRequest,
    ) -> Dict[str, Any]:
        match = re.search(
            r"^---\n.*", masking_request.reason, flags=re.MULTILINE | re.DOTALL
        )
        if not match:
            raise ValueError("Unable to find marker lines in masking request’s reason")
        return yaml.safe_load(match.group(0))

    @staticmethod
    def _get_requested_from_masking_request(
        requested: List[str],
    ) -> Iterator[Origin | ExtendedSWHID]:
        for s in requested:
            if s.startswith("swh:1:"):
                yield ExtendedSWHID.from_string(s)
            else:
                yield Origin(url=s)

    def handle_removal_notification_with_removal(
        self,
        notification_removal_identifier: str,
        secret_sharing: SecretSharing,
        recovery_bundle_path: str,
        ignore_requested=List[Origin | ExtendedSWHID],
        allow_empty_content_objects: bool = False,
        recompute_swhids_to_remove: bool = True,
    ) -> None:
        assert self.masking_admin is not None

        masking_request_slug = (
            f"{MASKING_REQUEST_IDENTIFIER_PREFIX}{notification_removal_identifier}"
        )
        # Recover data required to perform removal
        with self.masking_admin.conn:
            masking_request = self.masking_admin.find_request(masking_request_slug)
            if masking_request is None:
                raise MaskingRequestNotFound(masking_request_slug)
            info = self._get_info_from_masking_request(masking_request)
            requested = list(
                set(self._get_requested_from_masking_request(info["requested"]))
                - set(ignore_requested)
            )
            requested_swhids = [
                o.swhid() if isinstance(o, Origin) else o for o in requested
            ]
            # XXX should we add a flag to remove all of them?
            masking_states = self.masking_admin.get_states_for_request(
                masking_request.id
            )
            decision_pending_swhids = {
                swhid
                for swhid, state in masking_states.items()
                if state == MaskedState.DECISION_PENDING
            }

        # Perform removal
        if recompute_swhids_to_remove:
            try:
                removable = self.get_removable(requested_swhids)
            except RootsNotFound as e:
                _secho(
                    "Some requested objects were not found:",
                    err=True,
                    fg="red",
                    bold=True,
                )
                for label in e.get_labels(requested):
                    _secho(f"- {label}", err=True)
                raise
        else:
            removable = Removable(
                removable_swhids=list(decision_pending_swhids),
                referencing=requested_swhids,
            )

        decryption_key = self.create_recovery_bundle(
            secret_sharing=secret_sharing,
            requested=requested,
            removable=removable,
            recovery_bundle_path=recovery_bundle_path,
            removal_identifier=masking_request_slug,
            reason=info["reason"],
            allow_empty_content_objects=allow_empty_content_objects,
        )
        _secho(f"Recovery bundle decryption key: {decryption_key}", fg="blue")
        try:
            self.remove()
        except Exception as e:
            _secho(str(e), err=True, fg="red", bold=True)
            _secho("Rolling back…", fg="cyan")
            self.restore_recovery_bundle()
            raise

        # Clear removed objects in masking db
        recovery_bundle = RecoveryBundle(recovery_bundle_path)
        removed_swhids = recovery_bundle.swhids
        with self.masking_admin.conn:
            leftover_swhids = decision_pending_swhids - set(removed_swhids)
            record = f"Made {len(removed_swhids)} objects visible again after removal."
            if leftover_swhids:
                record += (
                    "\n\nObjects in state “decision pending” that were not removed:\n"
                )
                record += "\n".join(f"- {swhid}" for swhid in leftover_swhids)
            self.masking_admin.set_object_state(
                masking_request.id, MaskedState.VISIBLE, recovery_bundle.swhids
            )
            self.masking_admin.record_history(masking_request.id, record)

    def handle_removal_notification_by_changing_masked_status(
        self,
        notification_removal_identifier: str,
        masked_state: MaskedState,
    ):
        assert self.masking_admin is not None

        masking_request_slug = (
            f"{MASKING_REQUEST_IDENTIFIER_PREFIX}{notification_removal_identifier}"
        )
        with self.masking_admin.conn:
            masking_request = self.masking_admin.find_request(masking_request_slug)
            if masking_request is None:
                raise MaskingRequestNotFound(masking_request_slug)
            masking_states = self.masking_admin.get_states_for_request(
                masking_request.id
            )
            swhids = list(masking_states.keys())
            self.masking_admin.set_object_state(
                masking_request.id,
                masked_state,
                swhids,
            )
            match masked_state:
                case MaskedState.RESTRICTED:
                    masked_state_label = "permanently restricted"
                case MaskedState.VISIBLE:
                    masked_state_label = "visible again"
                case MaskedState.DECISION_PENDING:
                    masked_state_label = "restricted until a decision is made"
            record = f"Made {len(swhids)} objects {masked_state_label}."
            self.masking_admin.record_history(masking_request.id, record)



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import base64
import logging
import smtplib
import textwrap

import pytest

from swh.model.swhids import ExtendedSWHID
from swh.storage.proxies.masking.db import MaskedState, MaskingAdmin

from ..notifications import RemovalNotification


@pytest.fixture
def masking_admin(masking_db_postgresql_dsn) -> MaskingAdmin:
    return MaskingAdmin.connect(masking_db_postgresql_dsn)


def test_watch_notifications(
    mocker,
    notification_journal_writer,
    example_watcher,
    example_removal_notification,
):
    notification_journal_writer.write_additions(
        "removal_notification", [example_removal_notification]
    )
    process_removal_notification = mocker.patch.object(
        example_watcher, "process_removal_notification"
    )
    example_watcher.watch()
    process_removal_notification.assert_called_once_with(example_removal_notification)


def test_process_removal_notification(
    smtpd,
    example_watcher,
    masking_admin,
    example_removal_notification,
):
    example_watcher.process_removal_notification(example_removal_notification)

    masking_requests = masking_admin.get_requests(include_cleared_requests=True)
    assert len(masking_requests) == 1
    masking_request, masked_count = masking_requests[0]
    assert masked_count == 8
    assert (
        masking_request.slug == "removal-from-main-archive-example-removal-notification"
    )
    assert "https://example.com/swh/graph" in masking_request.reason
    assert (
        "swh:1:snp:0000000000000000000000000000000000000022" in masking_request.reason
    )
    states = masking_admin.get_states_for_request(masking_request.id)
    assert all(state == MaskedState.DECISION_PENDING for state in states.values())
    assert set(states.keys()) == {
        ExtendedSWHID.from_string(s)
        for s in (
            "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
            "swh:1:snp:0000000000000000000000000000000000000022",
            "swh:1:rel:0000000000000000000000000000000000000021",
            "swh:1:rev:0000000000000000000000000000000000000018",
            "swh:1:dir:0000000000000000000000000000000000000017",
            "swh:1:cnt:0000000000000000000000000000000000000015",
            "swh:1:cnt:0000000000000000000000000000000000000014",
            "swh:1:emd:101d70c3574c1e4b730d7ba8e83a4bdadc8691cb",
        )
    }
    assert len(smtpd.messages) == 1
    assert (
        "Removal from the main Software Heritage archive (example-removal-notification)"
        in smtpd.messages[0]["Subject"]
    )


def test_process_removal_notification_no_reason(
    smtpd,
    example_watcher,
    masking_admin,
    example_removal_notification,
):
    example_removal_notification.reason = None
    example_watcher.process_removal_notification(example_removal_notification)
    assert len(smtpd.messages) == 1

    msg = smtpd.messages[0]
    assert (
        "Removal from the main Software Heritage archive (example-removal-notification)"
        in msg["Subject"]
    )
    body = base64.b64decode(msg.get_payload()).decode()
    assert (
        "removed from the main Software Heritage archive "
        "for the following reason:\n\n\n\n"
    ) in body


def test_process_removal_notification_idempotency(
    example_watcher,
    masking_admin,
    example_removal_notification,
):
    example_watcher.process_removal_notification(example_removal_notification)
    example_watcher.process_removal_notification(example_removal_notification)

    assert len(masking_admin.get_requests(include_cleared_requests=True)) == 1


def test_process_removal_notification_failure(
    mocker,
    example_watcher,
    masking_admin,
    example_removal_notification,
):
    mocker.patch(
        "swh.alter.mirror_notification_watcher.MaskingAdmin.record_history",
        side_effect=IOError,
    )

    with pytest.raises(IOError):
        example_watcher.process_removal_notification(example_removal_notification)

    assert len(masking_admin.get_requests(include_cleared_requests=True)) == 0


def test_format_removal_email(
    example_watcher,
    example_removal_notification,
):
    text = example_watcher.format_removal_email(example_removal_notification)
    print(text)
    assert (
        textwrap.dedent(
            """\
        - origin: 1
        - snapshot: 1
        - release: 1
        - revision: 1
        - directory: 1
        - content: 2
        - raw extrinsic metadata: 1
        """
        )
        in text
    )
    assert "    We need to test stuff" in text
    assert "- https://example.com/swh/graph" in text
    assert "- swh:1:snp:0000000000000000000000000000000000000022" in text


def test_format_removal_email_with_missing_requested_objects(
    example_watcher,
    example_removal_notification,
):
    notification_d = example_removal_notification.to_dict()
    notification_d["requested"].append(
        "swh:1:snp:ffffffffffffffffffffffffffffffffffffffff"
    )
    notification_d["removed_objects"].append(
        "swh:1:snp:ffffffffffffffffffffffffffffffffffffffff"
    )
    text = example_watcher.format_removal_email(
        RemovalNotification.from_dict(notification_d)
    )
    assert (
        textwrap.dedent(
            """\
        - origin: 1
        - snapshot: 2
        - release: 1
        - revision: 1
        - directory: 1
        - content: 2
        - raw extrinsic metadata: 1
        """
        )
        in text
    )
    assert "- https://example.com/swh/graph" in text
    assert "- swh:1:snp:0000000000000000000000000000000000000022" in text
    assert (
        "- swh:1:snp:ffffffffffffffffffffffffffffffffffffffff ⚠️  Missing from mirror"
        in text
    )


def test_send_upstream_removal_email_no_recipients(
    caplog,
    mocker,
    example_watcher,
    example_removal_notification,
):
    mocker.patch(
        "smtplib.SMTP.sendmail",
        side_effect=smtplib.SMTPRecipientsRefused(recipients={}),
    )
    with caplog.at_level(logging.ERROR), pytest.raises(smtplib.SMTPRecipientsRefused):
        example_watcher.send_upstream_removal_email(example_removal_notification)
    assert "No recipients were specified" in caplog.text


def test_send_upstream_removal_email_all_recipients_refused(
    caplog,
    mocker,
    example_watcher,
    example_removal_notification,
):
    mocker.patch(
        "smtplib.SMTP.sendmail",
        side_effect=smtplib.SMTPRecipientsRefused(
            recipients={"!!!": (550, b"Nope"), "???": (550, b"Neither")}
        ),
    )
    with caplog.at_level(logging.ERROR), pytest.raises(smtplib.SMTPRecipientsRefused):
        example_watcher.send_upstream_removal_email(example_removal_notification)
    assert "All recipients were refused by the SMTP server" in caplog.text
    assert "- !!!: (550) Nope" in caplog.text
    assert "- ???: (550) Neither" in caplog.text


def test_send_upstream_removal_email_some_recipient_refused(
    caplog,
    mocker,
    example_watcher,
    example_removal_notification,
):
    mocker.patch(
        "smtplib.SMTP.sendmail",
        return_value={"two@example.org": (550, b"Recipient refused")},
    )
    with caplog.at_level(logging.WARNING):
        example_watcher.send_upstream_removal_email(example_removal_notification)
    assert "Partial error when sending email" in caplog.text
    assert "Some recipients were refused by the SMTP server" in caplog.text
    assert "- two@example.org: (550) Recipient refused" in caplog.text


def test_send_upstream_removal_email_refused_sender(
    caplog,
    mocker,
    example_watcher,
    example_removal_notification,
):
    mocker.patch(
        "smtplib.SMTP.sendmail",
        side_effect=smtplib.SMTPSenderRefused(
            code=550, msg=b"Sender not allowed", sender="test@example.org"
        ),
    )
    with caplog.at_level(logging.ERROR), pytest.raises(smtplib.SMTPSenderRefused):
        example_watcher.send_upstream_removal_email(example_removal_notification)
    assert "Sender “test@example.org” was refused" in caplog.text
    assert "(550) Sender not allowed" in caplog.text


def test_send_upstream_removal_email_error_response(
    caplog,
    mocker,
    example_watcher,
    example_removal_notification,
):
    mocker.patch(
        "smtplib.SMTP.sendmail",
        side_effect=smtplib.SMTPResponseException(
            code=500, msg=b"Unrecognized command"
        ),
    )
    with caplog.at_level(logging.ERROR), pytest.raises(smtplib.SMTPResponseException):
        example_watcher.send_upstream_removal_email(example_removal_notification)
    assert "(500) Unrecognized command" in caplog.text
